﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using UnityEngine.Networking;

public enum DamageType {
	Magical,
	Physical,
	Pure,
	Composite,
	HPRemoval,
	Universal
};


public enum CombatStat {
	None,
	Strength,
	Agility,
	Intelligence,
	Health,
	Mana,
	HealthRegen,
	ManaRegen,	
	AttackSpeed,
	AttackDamage,
	AttackRange,
	Armour,
	MagicResist,
	MoveSpeed,
	Experience,
	Gold
}

public enum Team {
	Neutral = 0,
	Blue = 1,
	Red = 2,
	Any = 3

}
[System.Serializable]
public class Modifier {
	public CombatStat stat;
	public float amount;
}

[System.Serializable]
public class Buf
{
	public Buf(CombatStat stat, float amount, float duration)
	{
		Modifier mod = new Modifier();
		mod.stat = stat;
		mod.amount = amount;
		mods.Add (mod);
		expireTimer = Time.time + duration;
	}
	public List<Modifier> mods = new List<Modifier>();
	public float expireTimer;
}

[System.Serializable]
public class Item
{

	public Item(CombatStat stat, float amount)
	{
		Buf b = new Buf(stat, amount, 0);
		bufs.Add (b);
	}
	public List<Buf> bufs = new List<Buf>();
}	

public partial class Combat : NetworkBehaviour {

	public delegate void FireDelegate(bool flag);
	public delegate void TakeDamageDelegate(int amount, Vector3 fromDirection, DamageType dt);
	public delegate void SetTargetDelegate(NetworkInstanceId id);
	public delegate void SetLevelDelegate(uint id);
	public delegate void GainXPDelegate(uint id);
	public delegate void AttackDelegate();
	public delegate void DieDelegate();
	public delegate void UseAbilityDelegate(int id, float cooldown);
	public delegate void AddBufDelegate(CombatStat stat, float amount, float duration);
	public delegate void AddItemDelegate(CombatStat stat, float amount);	
	
	[SyncEvent] public event FireDelegate EventFire;
	[SyncEvent] public event TakeDamageDelegate EventTakeDamage;
	[SyncEvent] public event SetTargetDelegate EventSetTarget;
	[SyncEvent] public event SetLevelDelegate EventSetLevel;
	[SyncEvent] public event GainXPDelegate EventGainXP;
	[SyncEvent] public event AttackDelegate EventAttack;
	[SyncEvent]	public event DieDelegate EventDie;
	[SyncEvent] public event UseAbilityDelegate EventUseAbility;
	[SyncEvent]	public event AddBufDelegate EventAddBuf;
	[SyncEvent]	public event AddItemDelegate EventAddItem;

	// Primary Attributes
	public float strength;
	public float agility;	
	public float intelligence;
	
	[SyncVar]
	public float maxHealth;
	
	[SyncVar]
	public float health;
	
	[SyncVar]
	public float maxMana;
	
	[SyncVar]
	public float mana;
	
	[SyncVar]
	public string combatName = "none";
	
	[SyncVar]
	public int experience = 0;
		
	[SyncVar]
	public Team team = Team.Blue;
	
	[SyncVar]
	public float gold = 625.0f;
	
	[SyncVar]
	public float currentLevel = 1;
		
	public float healthRegen;
	public float manaRegen;
	public float armor;
	public float attackDamage;
	public float damageReduction;
	public float magicResistance;
	public float attackRange;
	public bool invulnerable;
	public float moveSpeed;
	
	public float IAS; // increased attack speed
	public float attackTime;
	public float attacksPerSecond;
	
	float regenTimer;
	float attackTimer;
	public float[] abilityTimer = new float[4];
	public Combat target;
	bool isFiring;
	public bool isAlive = true;
	public bool isHero = false;
	
	// these would be per-hero
	const float strengthPerLevel = 2.2f;
	const float agilityPerLevel = 1.8f;
	const float intelligencePerLevel = 0.9f;
	const float baseDamage = 35.0f;
	const float baseAttackRange = 20.0f;
	const float abilityCooldown = 5.0f;
	const float baseMoveSpeed = 300.0f;
	
	const int abilityManaCost = 10;
	const float xpDistance = 50.0f;
	
	public CombatStat debugMod;
	
	public List<Buf> bufs = new List<Buf>();
	public List<Item> items = new List<Item>();
	
	void Start () {
		
		strength = 20;
		agility = 20;
		intelligence = 20;
		
		CalcStats();
		health = maxHealth;
		mana = maxMana;
		
		debugMod = CombatStat.None;
	}
	
	public override	void OnStartClient()
	{
		EventUseAbility += UseAbilityClient;
		EventAddBuf += AddBufClient;
		EventAddItem += AddItemClient;
	}
	
	
	public void AddItem(CombatStat stat, float amount)
	{
		Item item = new Item(stat, amount);
		items.Add (item);
		EventAddItem(stat, amount);
	}
	
	void AddItemClient(CombatStat stat, float amount)
	{
		if (NetworkServer.active) {
			// this is for localclient mode
			return;
		}
		Item item = new Item(stat, amount);
		items.Add (item);	
	}
		
	public void AddBuf(CombatStat stat, float amount, float duration)
	{
		Buf buf = new Buf(stat, amount, duration);
		bufs.Add (buf);
		EventAddBuf(stat, amount, duration);
	}
	
	void AddBufClient(CombatStat stat, float amount, float duration)
	{
		if (NetworkServer.active) {
			// this is for localclient mode
			return;
		}
		Buf buf = new Buf(stat, amount, duration);
		bufs.Add (buf);	
	}
		
	float GetBuf(CombatStat stat)
	{
		float amount = 0;
		foreach (Buf buf in bufs) {
			foreach(Modifier mod in buf.mods) {
				if (mod.stat == stat) {
					amount += mod.amount;
				}
			}
		}
		if (debugMod == stat) {
			amount += 400;
		}
		
		foreach (Item item in items) {
			foreach (Buf buf in item.bufs) {
				foreach(Modifier mod in buf.mods) {
					if (mod.stat == stat) {
						amount += mod.amount;
					}
				}
			}
		}
			
		return amount;
	}
	
	void CalcStats()
	{
		//todo: only once per frame?
		
		// FROM: http://dota2.gamepedia.com/Mechanics
		
		strength = 10 + strengthPerLevel * currentLevel + GetBuf(CombatStat.Strength);
		agility = 10 + agilityPerLevel * currentLevel + GetBuf(CombatStat.Agility);
		intelligence = 10 + intelligencePerLevel * currentLevel + GetBuf(CombatStat.Intelligence);
		
		maxHealth = 150 + strength * 19  + GetBuf(CombatStat.Health);
		healthRegen = 0.25f + 0.03f * strength + GetBuf(CombatStat.HealthRegen);
		if (healthRegen <= 0) {
			healthRegen = 0;
		}
		if (health > maxHealth) {
			health = maxHealth;
		}
		
		maxMana = intelligence * 13  + GetBuf(CombatStat.Mana);
		manaRegen = 0.01f + 0.04f * intelligence +  + GetBuf(CombatStat.ManaRegen);
		if (manaRegen <= 0) {
			manaRegen = 0;
		}
		if (mana > maxMana) {
			mana = maxMana;
		}	
		
		IAS = agility + GetBuf(CombatStat.AttackSpeed);
		
		attackTime = 1.7f / (1+IAS*0.01f);
		attacksPerSecond = (1 + IAS*0.01f) / 1.7f;
		
		// this should be primary attribute, not STR
		attackDamage = baseDamage + strength * 1 +  + GetBuf(CombatStat.AttackDamage);
		attackRange = baseAttackRange + GetBuf(CombatStat.AttackRange);
		
		armor = agility * 0.14f + GetBuf(CombatStat.Armour);
		damageReduction =  ((0.06f * armor) / (1 + 0.06f * armor));
		magicResistance = 0.25f + GetBuf(CombatStat.MagicResist);
		
		moveSpeed = baseMoveSpeed + GetBuf(CombatStat.MoveSpeed);
	}
	
	void Update () 
	{
		if (isServer)
		{
			if (regenTimer < Time.time) 
			{
				CalcStats ();
				regenTimer = Time.time + 1.0f;
				
				// health regen
				if (health < maxHealth) {
					health += healthRegen;
					if (health >= maxHealth) {
						health = maxHealth;
					}
				}
				
				// mana regen
				if (mana < maxMana) {
					mana += manaRegen;
					if (mana >= maxMana) {
						mana = maxMana;
					}
				}
				
				gold += 0.6f;
			}
			
			if (CanAttack() && isFiring)
			{
				DoAttack();
			}
		}		
		// check for expired bufs
		for (int i = bufs.Count - 1; i >= 0; i--)
		{ 
			if (Time.time > bufs[i].expireTimer) {
				bufs.RemoveAt(i);
			}
		}
	}
	
	public bool CanAttack()
	{
		return attackTimer < Time.time;
	}
	
	public void HasAttacked()
	{
		attackTimer = Time.time + attackTime;
	}
	
	void DoAttack()
	{		
		if (target == null) {
			return;
		}
		
		if (target.team == this.team) {
			return;
		}

		if (!IsTargetInRange(target.transform, attackRange)) {
			return;
		}
		
		if (!IsTargetInFront(target.transform, 45)) {
			return;
		}
		
		CalcStats();
		
		// generate attack event
		EventAttack();		
			
		//Debug.Log("Target " + target + " in range: " + range);
		if (!target.TakeDamage((int)attackDamage, transform.position, DamageType.Physical))
		{
		}
		HasAttacked();
	}
	
	bool UseMana(int amount)
	{
		if (mana < amount) {
			return false;
		}
		mana -= amount;
		if (mana < 0) {
			mana = 0;
		}
		return true;
	}
	
	public bool TakeDamage(int amount, Vector3 fromDirection, DamageType dt)
	{
		if (invulnerable) {
			return true;
		}
		
		if (!isAlive) {
			return false;
		}
		
		//Debug.Log ("TakeDamage " + gameObject + " amount=" + amount);
		
		float modifiedAmount = amount;
		if (dt == DamageType.Physical)
		{
			// apply armor
			modifiedAmount = amount - (amount * damageReduction);
		}
		else if (dt == DamageType.Magical)
		{
			// apply magic resist
			modifiedAmount = amount - (amount * magicResistance);
		}
		
		if (health < modifiedAmount)
		{
			// generate die event
			health = 0;
			isAlive = false;
			EventDie();
			//if (eDie != null) {
			//	eDie();
			//}
			
			foreach (Combat enemy in GetTargetsInRange(transform.position, xpDistance, Team.Any, team))
			{
				if (enemy.isHero) {
					enemy.GrantXP((int)currentLevel * 20 * 55);
				}
			}
			
			return false;
		}
		
		health -= modifiedAmount;
		// generate damage event
		EventTakeDamage((int)modifiedAmount, fromDirection, dt);
		return true;
	}
	
	public void AddGold(int amount)
	{
		gold += amount;
	}
	
	public bool SpendGold(int amount)
	{
		if (gold < amount){
			return false;
		}
		else
		{
			gold -= amount;
			return true;
		}
	}
	
	void GrantXP(int amount)
	{
		experience += amount;
		int xpForNextLevel = (int)(currentLevel+1)*100;
		while (experience >= xpForNextLevel)
		{
			experience -= xpForNextLevel;
			GainLevel();
			xpForNextLevel = (int)(currentLevel+1)*100;
		}
		EventGainXP((uint)amount);
	}
	
	void GainLevel()
	{
		currentLevel += 1;
		CalcStats();
		EventSetLevel((uint)currentLevel);
	}
	
	public void SetTarget(Combat newTarget)
	{
		if (newTarget != target) {
			// target event

			if (newTarget == null)
			{
				Debug.Log ("SetTarget " + gameObject + " set target to null");
				//XCallSetTargetDelegate(0);
				EventSetTarget(NetworkInstanceId.Invalid);
			} else {
				Debug.Log ("SetTarget " + gameObject + " set target to " + newTarget.gameObject);
				EventSetTarget(newTarget.GetComponent<NetworkIdentity>().netId); // TODO serialize object references
			}
		}
		target = newTarget;
	}
	
	public void StartFiring()
	{
		isFiring = true;
		EventFire(true);
	}
	
	public void StopFiring()
	{
		isFiring = false;
		EventFire(false);
	}

	public bool UseAbility(int id)
	{
		// use ability on current target
		if (abilityTimer[id] >= Time.time) {
			// ability not ready
			return false;
		}
		
		if (target == null) {
			return false;
		}
		
		// target must be in front of me
		if (target == this) {
			// no range check
		}
		else
		{
			if (!IsTargetInFront(target.transform, 60)) {
				return false;
			}
			
			// target must be in range
			if (!IsTargetInRange(target.transform, attackRange)) {
				return false;
			}
		}
		
		// check mana
		if (!UseMana (abilityManaCost)) {
			return false;
		}
				
		if (id == 3)
		{
			target.AddBuf(CombatStat.Strength, 20, 10);
		}
		else if (id == 2)
		{
			target.AddBuf(CombatStat.HealthRegen, 50, 5);
		}		
		else
		{
			// abilities should do different damage
			if (!target.TakeDamage(250, transform.position, DamageType.Magical))
			{
			}
		}
			
		// abilities should have different cooldowns
		abilityTimer[id] = Time.time + abilityCooldown;
		EventUseAbility(id, abilityCooldown);
				
		return true;
	}
	
	void UseAbilityClient(int id, float cooldown)
	{
		if (NetworkServer.active) {
			return;
		}
		abilityTimer[id] = Time.time + cooldown;
	}
	
	public Color GetTeamColor()
	{
		if (team == Team.Blue) {
			return Color.blue;
		} else if (team == Team.Red) {
			return Color.red;
		} else {
			return Color.grey;
		}
	}
		
	// ---------------- helper functions ----------------
	
	public bool IsTargetInRange(Transform target, float range)
	{
		Vector3 directionToTarget = target.position - transform.position;
		directionToTarget.y = 0; // distance in 2d
		float distance = directionToTarget.magnitude;
		return distance <= range;
	}
	
	public bool IsTargetInFront(Transform target, float frontAngle)
	{
		Vector3 directionToTarget = target.position - transform.position;
		directionToTarget.y = 0;
		float angle = Vector3.Angle(transform.forward, directionToTarget);
		return Mathf.Abs(angle) < frontAngle;
	}
	
	static public Combat[] GetTargetsInRange(Vector3 position, float range, Team matchTeam, Team excludeTeam)
	{
		List<Combat> outList = new List<Combat>();
		Collider[] hitColliders = Physics.OverlapSphere(position, range);
		foreach (Collider coll in hitColliders)
		{
			Combat c = coll.gameObject.GetComponent<Combat>();
			if (c == null) {
				continue;
			}
			if ((matchTeam == Team.Any || matchTeam == c.team))
			{
				if (excludeTeam == c.team)
				{
					//excluded
				}
				else
				{
					outList.Add (c);
				}
			}
		}
		return outList.ToArray();
	}
}
